/* mathlink.c	-- Daisuke Ikegami
 * Last modified: "Sun Jun 10 14:33:42 2001"
 */

#include "ruby.h"
#include "mathlink.h"

/** Mathematica::MathLink::MathLink
 ** abstract object that represents MathLink 
 **/ 

/*** eMathLinkError : subclass of Exception
 *** it represents MathLink Error
 ***/
VALUE eMathLinkError;

/*** mathlink and environment wrap MathLink variables ***/
struct mathlink 
{
  MLINK *link;
};

struct environment
{
  MLENV *env;
};

/*** This two macros are to get MathLink variables from Ruby objects ***/
#define GetMathLink(obj, linkp) {\
   if(NIL_P(obj)) {\
     rb_raise(rb_eRuntimeError, \
              "Cant get link obj. You maybe forget to open MathLink.");\
   }\
   Data_Get_Struct(obj, struct mathlink, linkp);\
   if(! linkp || ! linkp->link) {\
	 rb_raise(rb_eRuntimeError, \
			  "Cant get link pointer from the instance var link");\
   }\
}

#define GetEnvironment(obj, envp) {\
   if(NIL_P(obj)) {\
     rb_raise(rb_eRuntimeError, \
              "Cant get link obj. You maybe forget to open MathLink.");\
   }\
   Data_Get_Struct(obj, struct environment, envp);\
   if(! envp || ! envp->env) {\
	 rb_raise(rb_eRuntimeError, \
			  "Cant get env pointer from the instance var env");\
   }\
}

/*** deinitialize_environment(env) [private]
 *** deinitalizes MLENV env.
 *** this function is called by GC because we set this function
 *** on #initialize to deinitialize Environment
 *** See also --- MathLink#initialize
 ***/
static VALUE
deinitialize_environment(env)
	 VALUE env;
{
  struct environment *envp;

  GetEnvironment(env, envp);

  if(! *(envp->env)) {
	rb_raise(eMathLinkError, "Failed to deinitialize MathLink.");
  }

  MLDeinitialize(*(envp->env));
  DATA_PTR(env) = NULL;
  
  return Qnil;
}

/*** --- MathLink#initialize
 *** initializes MLENV env and make env an instance object.
 *** See also MLInitialize().
 ***/
static VALUE
mathlink_initialize(argc, argv, self)
	 int argc;
	 VALUE *argv, self;
{
  struct environment *envp;
  MLENV *env;
  VALUE obj, symbols, strings;

  env = ALLOC(MLENV);
  if(! (*env = MLInitialize(0))) {
	rb_raise(eMathLinkError,
			 "Failed to initialize MathLink. -- Failed MLInitialize(0)");
  }

  obj = Data_Make_Struct(rb_cObject, struct environment, 0,
						 deinitialize_environment, envp);
  envp->env = env;
  
  symbols = rb_hash_new();
  strings = rb_hash_new();
  
  rb_iv_set(self, "env", obj);
  rb_iv_set(self, "link", Qnil); /* set in --- MathLink#open */
  rb_iv_set(self, "symbols", symbols);
  rb_iv_set(self, "strings", strings);
  
  return self;
}

/*** --- MathLink#close
 *** closes MathLink.
 *** See also MLClose().
 ***/
static VALUE
mathlink_close(self)
	 VALUE self;
{
  struct mathlink *link;
  VALUE obj;

  obj = rb_iv_get(self, "link");
  GetMathLink(obj, link);

  /* reset hash table @symbols. */
  rb_iv_set(self, "strings", rb_hash_new());

  /* reset hash table @strings */
  rb_iv_set(self, "strings", rb_hash_new());
  
  MLClose(*(link->link));
  DATA_PTR(obj) = NULL;
  
  return Qnil;
}

/*** --- MathLink#activate
 *** activates MathLink.
 *** See also MLActivates().
 ***/
static VALUE
mathlink_activate(self)
	 VALUE self;
{
  struct mathlink *link;

  GetMathLink(rb_iv_get(self, "link"), link);
  
  MLActivate(*(link->link));
  
  return self;
}

/*** cant_get_envp [private]
 *** raises RuntimeError when we cannot get envp from the instance env.
 ***/
static void
cant_get_envp()
{
  rb_raise(eMathLinkError, "Failed to get environment status.");
  return;
}

/*** --- MathLink#open(argv)
 *** opens MathLink using some given options.
 *** it is assumed that argv is an array likes Ruby's constant ARGV.
 *** For example, you can write:
 ***   #! ruby
 ***   MathLink.new.open(ARGV.size, ARGV)
 ***   ...
 *** See also MLOpenArgv().
 ***/
static VALUE
mathlink_open(self, argc, argv)
	 VALUE self, argc, argv;
{
  int ml_argc, i;
  char **ml_argv;
  long errno;
  VALUE env, obj;
  struct environment *envp;
  struct mathlink *linkp;
  MLINK *link;

  ml_argc = NUM2INT(rb_Integer(argc));
  *ml_argv = ALLOCA_N(char, ml_argc);
  for(i=0; i < ml_argc; i++) {
	ml_argv[i] = STR2CSTR(rb_ary_shift(argv));
  }
  
  env = rb_iv_get(self, "env");
  Data_Get_Struct(env, struct environment, envp);
  if(! *(envp->env)) cant_get_envp();

  link = ALLOC(MLINK);
  *link = MLOpenArgv(*(envp->env), ml_argv, ml_argv+ml_argc, &errno);
  if(! *link) {
	rb_raise(eMathLinkError,
			 "Failed to open MathLink. %s: errno = %ld",
			 MLErrorString(*(envp->env), errno), errno);
  }

  obj = Data_Make_Struct(rb_cObject, struct mathlink,
						 0, mathlink_close, linkp);
  linkp->link = link;
  
  rb_iv_set(self, "link", obj);
  
  return self;
}

/*** --- MathLink#put_function(name, argc)
 *** puts a function with head given by a symbol with name and
 *** with argc arguments.
 *** See also MLPutFunction().
 ***/
static VALUE
mathlink_put_function(self, name, argc)
	 VALUE self, name, argc;
{
  struct mathlink *link;

  GetMathLink(rb_iv_get(self, "link"), link);
  MLPutFunction(*(link->link), STR2CSTR(rb_String(name)),
				NUM2INT(rb_Integer(argc)));

  return self;
}

/*** --MathLink#put_integer(num)
 *** puts the integer num.
 *** See also MLPutInteger().
 ***/
static VALUE
mathlink_put_integer(self, num)
	 VALUE self, num;
{
  struct mathlink *link;

  GetMathLink(rb_iv_get(self, "link"), link);
  MLPutInteger(*(link->link), NUM2INT(rb_Integer(num)));

  return self;
}

/*** --- MathLink#put_next(type)
 *** puts an object of the specified type.
 *** see also MLPutNext().
 ***/
static VALUE
mathlink_put_next(self, type)
	 VALUE self, type;
{
  struct mathlink *link;

  GetMathLink(rb_iv_get(self, "link"), link);
  MLPutNext(*(link->link), NUM2INT(rb_Integer(type)));

  return self;
}

/*** --- MathLink#put_byte_string
 *** puts a string containing only 8-bit character codes.
 *** see also MLPutByteString().
 ***/
static VALUE
mathlink_put_byte_string(self, str, len)
	 VALUE self, str, len;
{
  struct mathlink *link;

  GetMathLink(rb_iv_get(self, "link"), link);
  MLPutByteString(*(link->link),
				  STR2CSTR(rb_String(str)), NUM2LONG(rb_Integer(len)));

  return self;
}

/*** --- MathLink#get_integer
 *** gets an integer from the MathLink
 *** See also MLGetInteger().
 ***/
static VALUE
mathlink_get_integer(self)
	 VALUE self;
{
  struct mathlink *link;
  int result;
  VALUE obj;

  GetMathLink(rb_iv_get(self, "link"), link);
  MLGetInteger(*(link->link), &result);

  obj = INT2NUM(result);
  
  return obj;
}

/*** --- MathLink#get_real
 *** gets a floating-point number from the MathLink
 *** See also MLGetReal().
 ***/
static VALUE
mathlink_get_real(self)
	 VALUE self;
{
  struct mathlink *link;
  double real;
  VALUE obj;

  GetMathLink(rb_iv_get(self, "link"), link);
  MLGetReal(*(link->link), &real);

  obj = rb_float_new(real);
  
  return obj;
}

/*** --- MathLink#get_arg_count
 *** gets the number of arguments to a functon
 *** See also MLGetArgCount()
 ***/
static VALUE
mathlink_get_arg_count(self)
	 VALUE self;
{
  struct mathlink *link;
  long n;

  GetMathLink(rb_iv_get(self, "link"), link);
  MLGetArgCount(*(link->link), &n);

  return INT2NUM(n);
}

/*** --- MathLink#check_function(name)
 *** checks the number of the arguments of the function 'name'
 *** or return nil when 'name' is not a function or the name of the symbol
 *** does not match 'name'.
 *** See also MLCheckFunction().
 ***/
static VALUE
mathlink_check_function(self, name)
	 VALUE self, name;
{
  struct mathlink *link;
  long n;

  GetMathLink(rb_iv_get(self, "link"), link);

  if(! MLCheckFunction(*(link->link), STR2CSTR(rb_String(name)), &n)) {
	return Qnil;
  } else {
	return INT2NUM(n);
  }
}

/*** --- MathLink#end_packet
 *** specifies that a packet expression is complete and
 *** is ready to be sent on the specified link.
 *** See also MLEndPacket().
 ***/
static VALUE
mathlink_end_packet(self)
	 VALUE self;
{
  struct mathlink *link;

  GetMathLink(rb_iv_get(self, "link"), link);

  MLEndPacket(*(link->link));

  return self;
}

/*** --- MathLink#get_next
 *** goes to the next object on link and gets its type.
 *** See also MLGetNext().
 ***/
static VALUE
mathlink_get_next(self)
	 VALUE self;
{
  struct mathlink *link;

  GetMathLink(rb_iv_get(self, "link"), link);

  return INT2NUM(MLGetNext(*(link->link)));
}

/*** --- MathLink#disown_symbol(symbol)
 *** disowns memory allocated by MLGetSymbol or MLGetFunction
 *** to store the character string s corresponding to the name of
 *** a symbol.
 *** See also MLDisownSymbol().
 ***/
static VALUE
mathlink_disown_symbol(self, symbol)
	 VALUE self, symbol;
{
  char *name;
  struct mathlink *link;

  name = STR2CSTR(rb_String(symbol));
  
  /* delete symbol `self' on the hash table `symbols' */
  rb_funcall(rb_iv_get(self, "symbols"), rb_intern("delete"),
			 1, rb_str_new2(name));
  
  GetMathLink(rb_iv_get(self, "link"), link);
  MLDisownSymbol(*(link->link), name);
  
  return Qnil;
}

/*** --- MathLink#get_symbol
 *** gets the name of a symbol from the MathLink connection.
 *** See also MLGetSymbol().
 ***/
static VALUE
mathlink_get_symbol(self)
	 VALUE self;
{
  char *name;
  struct mathlink *link;
  VALUE result;
  
  GetMathLink(rb_iv_get(self, "link"), link);

  /* symbol is allocated by Mathematica kernel.
	 So we dont need to ALLOC(symbol). */
  MLGetSymbol(*(link->link), (kcharpp_ct) &name);

  if(! name) {
	result = Qnil;
  } else {
	result = rb_str_new2(name);
	/* store symbols(hash) the name to disown memory allocated by MathLink */
	rb_hash_aset(rb_iv_get(self, "symbols"), result, Qtrue);
  }
  
  return result;
}

/*** --- MathLink#disown_string(str)
 *** disowns memory allocated by MLGetString() to store the character str
 *** See also MLDisownString().
 ***/
static VALUE
mathlink_disown_string(self, str)
	 VALUE self, str;
{
  char *name;
  struct mathlink *link;

  name = STR2CSTR(rb_String(str));
  
  /* delete symbol `self' on the hash table `symbols' */
  rb_funcall(rb_iv_get(self, "strings"), rb_intern("delete"),
			 1, rb_str_new2(name));
  
  GetMathLink(rb_iv_get(self, "link"), link);
  MLDisownString(*(link->link), name);
  
  return Qnil;
}

/*** --- MathLink#get_string
 *** gets a character string from the MathLink connection
 *** See also MLGetString().
 ***/
static VALUE
mathlink_get_string(self)
	 VALUE self;
{
  char *str;
  struct mathlink *link;
  VALUE result;
  
  GetMathLink(rb_iv_get(self, "link"), link);

  /* string is allocated by Mathematica kernel.
	 So we dont need to ALLOC(str). */
  MLGetString(*(link->link), (kcharpp_ct) &str);

  if(! str) {
	result = Qnil;
  } else {
	result = rb_str_new2(str);
	/* store strings(hash) the name to disown memory allocated by MathLink */
	rb_hash_aset(rb_iv_get(self, "strings"), result, Qtrue);
  }
  
  return result;
}

/*** --- MathLink#next_packet
 *** goes to the next packet on link and returns a constant to
 *** indicate its head.
 *** See also MLNextPacket().
 ***/
static VALUE
mathlink_next_packet(self)
	 VALUE self;
{
  struct mathlink *link;
  
  GetMathLink(rb_iv_get(self, "link"), link);

  return INT2NUM(MLNextPacket(*(link->link)));
}

/*** --- MathLink#new_packet
 *** skips to the end of the current packet on link.
 *** See also MLNewPacket().
 ***/
static VALUE
mathlink_new_packet(self)
	 VALUE self;
{
  struct mathlink *link;

  GetMathLink(rb_iv_get(self, "link"), link);

  MLNewPacket(*(link->link));

  return self;
}

/*** --- MathLink#error
 *** returns an error number or nil if there is no error.
 *** See also MLError().
 ***/
static VALUE
mathlink_error(self)
	 VALUE self;
{
  int error;
  struct mathlink *link;
  
  GetMathLink(rb_iv_get(self, "link"), link);

  error = MLError(*(link->link));

  if(error)
	return INT2NUM(error);
  else
	return Qnil;
}

/*** --- MathLink#error_message
 *** returns a character string describing the last error to occur on link.
 *** See also MLErrorMessage().
 ***/
static VALUE
mathlink_error_message(self)
	 VALUE self;
{
  struct mathlink *link;
  
  GetMathLink(rb_iv_get(self, "link"), link);

  return rb_str_new2(MLErrorMessage(*(link->link)));
}

void
Init_mathlink()
{
  VALUE mMathematica, mMathLink, cMathLink;
  
  mMathematica = rb_define_module("Mathematica");
  mMathLink = rb_define_module_under(mMathematica, "MathLink");

  cMathLink = rb_define_class_under(mMathLink, "MathLink", rb_cObject);

  eMathLinkError = rb_define_class("MathLinkError", rb_eStandardError);
  
  rb_define_method(cMathLink, "initialize", mathlink_initialize, -1);

  rb_define_method(cMathLink, "open", mathlink_open, 2);
  rb_define_method(cMathLink, "close", mathlink_close, 0);
  rb_define_method(cMathLink, "activate", mathlink_activate, 0);

  rb_define_method(cMathLink, "put_function", mathlink_put_function, 2);
  rb_define_method(cMathLink, "put_integer", mathlink_put_integer, 1);
  rb_define_method(cMathLink, "put_next", mathlink_put_next, 1);
  rb_define_method(cMathLink, "put_byte_string", mathlink_put_byte_string, 2);
  
  rb_define_method(cMathLink, "get_next", mathlink_get_next, 0);
  rb_define_method(cMathLink, "get_symbol", mathlink_get_symbol, 0);
  rb_define_method(cMathLink, "disown_symbol", mathlink_disown_symbol, 1);
  rb_define_method(cMathLink, "get_string", mathlink_get_string, 0);
  rb_define_method(cMathLink, "disown_string", mathlink_disown_string, 1);
  rb_define_method(cMathLink, "get_integer", mathlink_get_integer, 0);
  rb_define_method(cMathLink, "get_real", mathlink_get_real, 0);

  rb_define_method(cMathLink, "get_arg_count", mathlink_get_arg_count, 0);
  
  rb_define_method(cMathLink, "check_function", mathlink_check_function, 1);

  rb_define_method(cMathLink, "new_packet", mathlink_new_packet, 0);
  rb_define_method(cMathLink, "next_packet", mathlink_next_packet, 0);
  rb_define_method(cMathLink, "end_packet", mathlink_end_packet, 0);

  rb_define_method(cMathLink, "error", mathlink_error, 0);
  rb_define_method(cMathLink, "error_message", mathlink_error_message, 0);  

  /* these global variable is prepered by mprep. 
	 I guess we dont need to set. See also Mathematica Manual about MLAbort.
	 rb_define_const(mMathLink, "MLAbort", INT2NUM(MLAbort));
	 rb_define_const(mMathLink, "MLDone", INT2NUM(MLDone));
  */
  
  rb_define_const(mMathLink, "RETURNPKT", INT2NUM(RETURNPKT));

  /*
	rb_define_const(mMathLink, "MLE", INT2NUM(MLE));
  */
  rb_define_const(mMathLink, "MLEUNKNOWN", INT2NUM(MLEUNKNOWN));
  rb_define_const(mMathLink, "MLEOK", INT2NUM(MLEOK));
  rb_define_const(mMathLink, "MLEDEAD", INT2NUM(MLEDEAD));
  rb_define_const(mMathLink, "MLEGBAD", INT2NUM(MLEGBAD));
  rb_define_const(mMathLink, "MLEGSEQ", INT2NUM(MLEGSEQ));
  rb_define_const(mMathLink, "MLEPBTK", INT2NUM(MLEPBTK));
  rb_define_const(mMathLink, "MLEPSEQ", INT2NUM(MLEPSEQ));
  rb_define_const(mMathLink, "MLEPBIG", INT2NUM(MLEPBIG));
  rb_define_const(mMathLink, "MLEOVFL", INT2NUM(MLEOVFL));
  rb_define_const(mMathLink, "MLEMEM", INT2NUM(MLEMEM));
  rb_define_const(mMathLink, "MLEACCEPT", INT2NUM(MLEACCEPT));
  rb_define_const(mMathLink, "MLECONNECT", INT2NUM(MLECONNECT));
  rb_define_const(mMathLink, "MLECLOSED", INT2NUM(MLECLOSED));
  rb_define_const(mMathLink, "MLEDEPTH", INT2NUM(MLEDEPTH));
  rb_define_const(mMathLink, "MLENODUPFCN", INT2NUM(MLENODUPFCN));
  rb_define_const(mMathLink, "MLENOACK", INT2NUM(MLENOACK));
  rb_define_const(mMathLink, "MLENODATA", INT2NUM(MLENODATA));
  rb_define_const(mMathLink, "MLENOTDELIVERED", INT2NUM(MLENOTDELIVERED));
  rb_define_const(mMathLink, "MLENOMSG", INT2NUM(MLENOMSG));
  rb_define_const(mMathLink, "MLEFAILED", INT2NUM(MLEFAILED));
  rb_define_const(mMathLink, "MLEGETENDEXPR", INT2NUM(MLEGETENDEXPR));
  rb_define_const(mMathLink, "MLEPUTENDPACKET", INT2NUM(MLEPUTENDPACKET));
  rb_define_const(mMathLink, "MLENEXTPACKET", INT2NUM(MLENEXTPACKET));
  rb_define_const(mMathLink, "MLEUNKNOWNPACKET", INT2NUM(MLEUNKNOWNPACKET));
  rb_define_const(mMathLink, "MLEGETENDPACKET", INT2NUM(MLEGETENDPACKET));
  rb_define_const(mMathLink, "MLEABORT", INT2NUM(MLEABORT));
  rb_define_const(mMathLink, "MLEMORE", INT2NUM(MLEMORE));
  rb_define_const(mMathLink, "MLENEWLIB", INT2NUM(MLENEWLIB));
  rb_define_const(mMathLink, "MLEOLDLIB", INT2NUM(MLEOLDLIB));
  rb_define_const(mMathLink, "MLEBADPARAM", INT2NUM(MLEBADPARAM));
  rb_define_const(mMathLink, "MLENOTIMPLEMENTED", INT2NUM(MLENOTIMPLEMENTED));
  rb_define_const(mMathLink, "MLEINIT", INT2NUM(MLEINIT));
  rb_define_const(mMathLink, "MLEARGV", INT2NUM(MLEARGV));
  rb_define_const(mMathLink, "MLEPROTOCOL", INT2NUM(MLEPROTOCOL));
  rb_define_const(mMathLink, "MLEMODE", INT2NUM(MLEMODE));
  rb_define_const(mMathLink, "MLELAUNCH", INT2NUM(MLELAUNCH));
  rb_define_const(mMathLink, "MLELAUNCHAGAIN", INT2NUM(MLELAUNCHAGAIN));
  rb_define_const(mMathLink, "MLELAUNCHSPACE", INT2NUM(MLELAUNCHSPACE));
  rb_define_const(mMathLink, "MLENOPARENT", INT2NUM(MLENOPARENT));
  rb_define_const(mMathLink, "MLENAMETAKEN", INT2NUM(MLENAMETAKEN));
  rb_define_const(mMathLink, "MLENOLISTEN", INT2NUM(MLENOLISTEN));
  rb_define_const(mMathLink, "MLEBADNAME", INT2NUM(MLEBADNAME));
  rb_define_const(mMathLink, "MLEBADHOST", INT2NUM(MLEBADHOST));
  rb_define_const(mMathLink, "MLERESOURCE", INT2NUM(MLERESOURCE));
  rb_define_const(mMathLink, "MLELAUNCHFAILED", INT2NUM(MLELAUNCHFAILED));
  rb_define_const(mMathLink, "MLELAUNCHNAME", INT2NUM(MLELAUNCHNAME));
  rb_define_const(mMathLink, "MLELAST", INT2NUM(MLELAST));
  rb_define_const(mMathLink, "MLETRACEON", INT2NUM(MLETRACEON));
  rb_define_const(mMathLink, "MLETRACEOFF", INT2NUM(MLETRACEOFF));
  rb_define_const(mMathLink, "MLEDEBUG", INT2NUM(MLEDEBUG));
  rb_define_const(mMathLink, "MLEASSERT", INT2NUM(MLEASSERT));
  rb_define_const(mMathLink, "MLEUSER", INT2NUM(MLEUSER));

  /*
	rb_define_const(mMathLink, "MLTK", INT2NUM(MLTK));
  */
  rb_define_const(mMathLink, "MLTKOLDINT", INT2NUM(MLTKOLDINT));
  rb_define_const(mMathLink, "MLTKOLDREAL", INT2NUM(MLTKOLDREAL));
  rb_define_const(mMathLink, "MLTKFUNC", INT2NUM(MLTKFUNC));
  rb_define_const(mMathLink, "MLTKERROR", INT2NUM(MLTKERROR));
  rb_define_const(mMathLink, "MLTKERR", INT2NUM(MLTKERR));
  rb_define_const(mMathLink, "MLTKSTR", INT2NUM(MLTKSTR));
  rb_define_const(mMathLink, "MLTKSYM", INT2NUM(MLTKSYM));
  rb_define_const(mMathLink, "MLTKREAL", INT2NUM(MLTKREAL));
  rb_define_const(mMathLink, "MLTKINT", INT2NUM(MLTKINT));
  rb_define_const(mMathLink, "MLTKPCTEND", INT2NUM(MLTKPCTEND));
  rb_define_const(mMathLink, "MLTKAPCTEND", INT2NUM(MLTKAPCTEND));
  rb_define_const(mMathLink, "MLTKEND", INT2NUM(MLTKEND));
  rb_define_const(mMathLink, "MLTKAEND", INT2NUM(MLTKAEND));
  rb_define_const(mMathLink, "MLTKSEND", INT2NUM(MLTKSEND));
  rb_define_const(mMathLink, "MLTKCONT", INT2NUM(MLTKCONT));
  rb_define_const(mMathLink, "MLTKELEN", INT2NUM(MLTKELEN));
  rb_define_const(mMathLink, "MLTKNULL", INT2NUM(MLTKNULL));
  rb_define_const(mMathLink, "MLTKOLDSYM", INT2NUM(MLTKOLDSYM));
  rb_define_const(mMathLink, "MLTKOLDSTR", INT2NUM(MLTKOLDSTR));
  rb_define_const(mMathLink, "MLTKPACKED", INT2NUM(MLTKPACKED));
  rb_define_const(mMathLink, "MLTKARRAY", INT2NUM(MLTKARRAY));
  rb_define_const(mMathLink, "MLTKDIM", INT2NUM(MLTKDIM));
  rb_define_const(mMathLink, "MLLENGTH_DECODER", INT2NUM(MLLENGTH_DECODER));
  rb_define_const(mMathLink, "MLTKPACKED_DECODER", INT2NUM(MLTKPACKED_DECODER));
  rb_define_const(mMathLink, "MLTKARRAY_DECODER", INT2NUM(MLTKARRAY_DECODER));
  rb_define_const(mMathLink, "MLTKMODERNCHARS_DECODER", INT2NUM(MLTKMODERNCHARS_DECODER));
  rb_define_const(mMathLink, "MLTKNULLSEQUENCE_DECODER", INT2NUM(MLTKNULLSEQUENCE_DECODER));
  rb_define_const(mMathLink, "MLTKALL_DECODERS", INT2NUM(MLTKALL_DECODERS));
  rb_define_const(mMathLink, "MLTK_FIRSTUSER", INT2NUM(MLTK_FIRSTUSER));
  rb_define_const(mMathLink, "MLTK_LASTUSER", INT2NUM(MLTK_LASTUSER));
  
  return;
}
